//! # Process
//!
//! $$
//! dX_t=a(t,X_t)dt+b(t,X_t)dW_t
//! $$
//!
pub mod bm;
pub mod cbms;
pub mod ccustom;
pub mod cfbms;
pub mod cpoisson;
pub mod customjt;
pub mod fbm;
pub mod lfsm;
pub mod poisson;
