//! # Jump
//!
//! $$
//! dX_t=a(t,X_t)dt+b(t,X_t)dW_t
//! $$
//!
pub mod bates;
pub mod cgmy;
pub mod cts;
pub mod ig;
pub mod jump_fou;
pub mod jump_fou_custom;
pub mod kou;
pub mod levy_diffusion;
pub mod merton;
pub mod nig;
pub mod rdts;
pub mod vg;
